#coding: UTF8
"""
mailer module

Simple front end to the smtplib and email modules,
to simplify sending email.

A lot of this code was taken from the online examples in the
email module documentation:
http://docs.python.org/library/email-examples.html

Released under MIT license.

Version 0.5 is based on a patch by Douglas Mayle

Sample code:

    import mailer

    message = mailer.Message()
    message.From = "me@example.com"
    message.To = "you@example.com"
    message.RTo = "you@example.com"
    message.Subject = "My Vacation"
    message.Body = open("letter.txt", "rb").read()
    message.attach("picture.jpg")

    sender = mailer.Mailer('mail.example.com')
    sender.send(message)

"""

import smtplib
import socket
import threading
import queue
import uuid

# this is to support name changes
# from version 2.4 to version 2.5
try:
    from email import encoders
    from email.header import make_header
    from email.mime.audio import MIMEAudio
    from email.mime.base import MIMEBase
    from email.mime.image import MIMEImage
    from email.mime.multipart import MIMEMultipart
    from email.mime.text import MIMEText
except ImportError:
    from email import Encoders as encoders
    from email.Header import make_header
    from email.MIMEAudio import MIMEAudio
    from email.MIMEBase import MIMEBase
    from email.MIMEImage import MIMEImage
    from email.MIMEMultipart import MIMEMultipart
    from email.MIMEText import MIMEText

# For guessing MIME type based on file name extension
import mimetypes
import time

from os import path

__version__ = "0.8.1"
__author__ = "Ryan Ginstrom"
__license__ = "MIT"
__description__ = "A module to send email simply in Python"


class Mailer(object):
    """
    Represents an SMTP connection.

    Use login() to log in with a username and password.
    """

    def __init__(self, host="localhost", port=0, use_tls=False, usr=None, pwd=None, use_ssl=False, use_plain_auth=False,
                 timeout=socket._GLOBAL_DEFAULT_TIMEOUT):
        self.host           = host
        self.port           = port
        self.use_tls        = use_tls
        self.use_ssl        = use_ssl
        self.use_plain_auth = use_plain_auth
        self._usr           = usr
        self._pwd           = pwd
        self.timeout        = timeout

    def login(self, usr, pwd):
        self._usr = usr
        self._pwd = pwd

    def send(self, msg, debug=False):
        """
        Send one message or a sequence of messages.

        Every time you call send, the mailer creates a new
        connection, so if you have several emails to send, pass
        them as a list:
        mailer.send([msg1, msg2, msg3])
        """
        if self.use_ssl:
            server = smtplib.SMTP_SSL(self.host, self.port, timeout=self.timeout)
        else:
            server = smtplib.SMTP(self.host, self.port, timeout=self.timeout)

        if debug:
            server.set_debuglevel(1)

        if self._usr and self._pwd:
            if self.use_tls is True:
                server.ehlo()
                server.starttls()
                server.ehlo()

            if self.use_plain_auth is True:
                server.esmtp_features["auth"] = "LOGIN PLAIN"

            server.login(self._usr, self._pwd)

        if isinstance(msg, Message):
            msg = [msg]

        for m in msg:
            self._send(server, m)

        server.quit()

    def _send(self, server, msg):
        """
        Sends a single message using the server
        we created in send()
        """
        me = msg.From
        if isinstance(msg.To, str):
            to = [msg.To]
        else:
            to = list(msg.To)

        cc = []
        if msg.CC:
            if isinstance(msg.CC, str):
                cc = [msg.CC]
            else:
                cc = list(msg.CC)

        bcc = []
        if msg.BCC:
            if isinstance(msg.BCC, str):
                bcc = [msg.BCC]
            else:
                bcc = list(msg.BCC)

        rto = []
        if msg.RTo:
            if isinstance(msg.RTo, str):
                rto = [msg.RTo]
            else:
                rto = list(msg.RTo)

        you = to + cc + bcc
        server.sendmail(me, you, msg.as_string())


class Message(object):
    """
    Represents an email message.

    Set the To, From, Reply-To, Subject, and Body attributes as plain-text strings.
    Optionally, set the Html attribute to send an HTML email, or use the
    attach() method to attach files.

    Use the charset property to send messages using other than us-ascii

    If you specify an attachments argument, it should be a list of
    attachment filenames: ["file1.txt", "file2.txt"]

    `To` should be a string for a single address, and a sequence
    of strings for multiple recipients (castable to list)

    Send using the Mailer class.
    """

    def __init__(self, **kwargs):
        """
        Parameters and default values (parameter names are case insensitive):
            To=None, From=None, RTo=None, CC=None, BCC=None, Subject=None, Body=None, Html=None,
            Date=None, Attachments=None, Charset=None, Headers=None
        """

        # extract  parameters and convert names to lowercase
        params = {}
        for i in kwargs:
            params[i.lower()] = kwargs[i]

        # preprocess attachments
        self.attachments = []
        attachments = params.get('attachments', None)
        if attachments:
            for attachment in attachments:
                if isinstance(attachment, str):
                    self.attachments.append((attachment, None, None, None, None))
                else:
                    try:
                        length = len(attachment)
                    except TypeError:
                        length = None
                    else:
                        if length is None or length <= 4:
                            self.attachments.append((attachment, None, None, None, None))
                        else:
                            self.attachments.append((tuple(attachment) + (None, None, None, None))[:4])


        self.To         = params.get('to', None)
        self.RTo        = params.get('rto', None)
        self.CC         = params.get('cc', None)
        self.BCC        = params.get('bcc', None)
        self.From       = params.get('from', None) # string or iterable
        self.Subject    = params.get('subject', '') # string
        self.Body       = params.get('body', None)
        self.Html       = params.get('html', None)
        self.Date       = params.get('date', time.strftime("%a, %d %b %Y %H:%M:%S %z", time.gmtime()))
        self.charset    = params.get('charset', 'us-ascii')
        self.Headers    = params.get('headers', {})

        if isinstance(self.Body, str):
            self.Body = self.Body.encode(self.charset)

        self.message_id = self.make_key()


    def make_key(self):
        return str(uuid.uuid4())

    def header(self, key, value):
        self.Headers[key] = value

    def as_string(self):
        """Get the email as a string to send in the mailer"""

        if not self.attachments:
            return self._plaintext()
        else:
            return self._multipart()

    def _plaintext(self):
        """Plain text email with no attachments"""

        if not self.Html:
            msg = MIMEText(self.Body, 'plain', self.charset)
        else:
            msg = self._with_html()

        self._set_info(msg)
        return msg.as_string()

    def _with_html(self):
        """There's an html part"""

        outer = MIMEMultipart('alternative')

        part1 = MIMEText(self.Body, 'plain', self.charset)
        part2 = MIMEText(self.Html, 'html', self.charset)

        outer.attach(part1)
        outer.attach(part2)

        return outer

    def _set_info(self, msg):
        if self.charset == 'us-ascii':
            msg['Subject'] = self.Subject
            msg['From'] = self.From

        else:
            if isinstance(self.Subject, str):
                subject = self.Subject
            else:
                subject = str(self.Subject, self.charset)
            msg['Subject'] = str(make_header([(subject, self.charset)]))

            if isinstance(self.From, str):
                from_ = self.From
            else:
                from_ = str(self.From, self.charset)
            msg['From'] = str(make_header([(from_, self.charset)]))


        if isinstance(self.To, str):
            msg['To'] = self.To
        else:
            self.To = list(self.To)
            msg['To'] = ", ".join(self.To)

        if self.RTo:
            if isinstance(self.RTo, str):
                msg.add_header('reply-to', self.RTo)
            else:
                self.RTo = list(self.RTo)
                msg.add_header('reply-to', ", ".join(self.RTo))

        if self.CC:
            if isinstance(self.CC, str):
                msg['CC'] = self.CC
            else:
                self.CC = list(self.CC)
                msg['CC'] = ", ".join(self.CC)


        if self.BCC:
            if isinstance(self.BCC, str):
                msg['BCC'] = self.BCC
            else:
                self.BCC = list(self.BCC)
                msg['BCC'] = ", ".join(self.BCC)


        if self.Headers:
            for key, value in list(self.Headers.items()):
                msg[key] = str(value).encode(self.charset)


        msg['Date'] = self.Date

    def _multipart(self):
        """The email has attachments"""

        msg = MIMEMultipart('related')

        if self.Html:
            outer = MIMEMultipart('alternative')

            part1 = MIMEText(self.Body, 'plain', self.charset)
            part1.add_header('Content-Disposition', 'inline')

            part2 = MIMEText(self.Html, 'html', self.charset)
            part2.add_header('Content-Disposition', 'inline')

            outer.attach(part1)
            outer.attach(part2)
            msg.attach(outer)
        else:
            msg.attach(MIMEText(self.Body, 'plain', self.charset))

        self._set_info(msg)
        msg.preamble = self.Subject

        for filename, cid, mimetype, content, charset in self.attachments:
            self._add_attachment(msg, filename, cid, mimetype, content, charset)

        return msg.as_string()

    def _add_attachment(self, outer, filename, cid, mimetype, content, charset):
        """
        If mimetype is None, it will try to guess the mimetype
        """
        if mimetype:
            ctype = mimetype
            encoding = None
        else:
            ctype, encoding = mimetypes.guess_type(filename)

        if ctype is None or encoding is not None:
            # No guess could be made, or the file is encoded (compressed), so
            # use a generic bag-of-bits type.
            ctype = 'application/octet-stream'

        maintype, subtype = ctype.split('/', 1)
        if not content:
            with open(filename, 'rb') as fp:
                content = fp.read()

        if maintype == 'text':
            # Note: we should handle calculating the charset
            msg = MIMEText(content, _subtype=subtype, _charset=charset)
        elif maintype == 'image':
            msg = MIMEImage(content, _subtype=subtype)
        elif maintype == 'audio':
            msg = MIMEAudio(content, _subtype=subtype)
        else:
            msg = MIMEBase(maintype, subtype)
            msg.set_payload(content)
            # Encode the payload using Base64
            encoders.encode_base64(msg)

        # Set the content-ID header
        if cid:
            msg.add_header('Content-ID', '<%s>' % cid)
            msg.add_header('Content-Disposition', 'inline')
        else:
            # Set the filename parameter
            msg.add_header('Content-Disposition', 'attachment', filename=path.basename(filename))

        outer.attach(msg)

    def attach(self, filename, cid=None, mimetype=None, content=None, charset=None):
        """
        Attach a file to the email. Specify the name of the file;
        Message will figure out the MIME type and load the file.

        Specify mimetype to set the MIME type manually. The content
        argument take the contents of the file if they are already loaded
        in memory.
        """

        self.attachments.append((filename, cid, mimetype, content, charset))


class Manager(threading.Thread):
    """
    Manages the sending of email in the background.

    You can supply it with an instance of class Mailer or pass in the same
    parameters that you would have used to create an instance of Mailer.

    If a message was succesfully sent, self.results[msg.message_id] returns a 3
    element tuple (True/False, err_code, err_message).
    """

    def __init__(self, mailer=None, callback=None, **kwargs):
        threading.Thread.__init__(self)

        self.queue = queue.Queue()
        self.mailer = mailer
        self.abort = False
        self.callback = callback
        self._results = {}
        self._result_lock = threading.RLock()

        if self.mailer is None:
            self.mailer = Mailer(
                host=kwargs.get('host', 'localhost'),
                port=kwargs.get('port', 25),
                use_tls=kwargs.get('use_tls', False),
                usr=kwargs.get('usr', None),
                pwd=kwargs.get('pwd', None),
            )

    def __getattr__(self, name):
        if name == 'results':
            with self._result_lock:
                return self._results
        else:
            return None

    def run(self):

        while self.abort is False:
            msg = self.queue.get(block=True)
            if msg is None:
                break

            if isinstance(msg, Message):
                msg = [msg]

            for m in msg:
                try:
                    self.results[m.message_id] = (False, -1, '')
                    self.mailer.send(m)
                    self.results[m.message_id] = (True, 0, '')

                except Exception as e:
                    args = e.args
                    if len(args) < 2:
                        args = (-1, e.args[0])

                    self.results[m.message_id] = (False, args[0], args[1])

                if self.callback:
                    try:
                        self.callback(m.message_id)
                    except Exception:
                        pass

            # endfor

            self.queue.task_done()

    def send(self, msg):
        self.queue.put(msg)



import pathlib
import pickle
from typing import Union, Any, Optional
import sys
from datetime import datetime, date
import re
import os
import math
from collections import defaultdict
import json
import yaml
import inspect
from dateutil import parser

from itertools import chain, islice

from divinegift import logger
from divinegift.errors import NoFileNamePassedError
from divinegift import version
#########################################################################


datetime_regex = r'20\d{2}(-|\/)((0[1-9])|(1[0-2]))(-|\/)((0[1-9])|([1-2][0-9])|(3[0-1]))(T|\s)(([0-1][0-9])|(2[0-3])):([0-5][0-9]):([0-5][0-9])'


def check_dg_version():
    import requests
    r = requests.get('https://pypi.org/pypi/divinegift/json')
    ver = r.json().get('info').get('version')
    if ver != version:
        print(f"Your version of divinegift is outdated. Please update it by command 'pip3 install divinegift=={ver}'")


def dict_compare(old_dict: dict, new_dict: dict):
    """
    Compare 2 dict
    :param old_dict: First dict to compare
    :param new_dict: Second dict
    :return: Dict with results of comparing
    """
    old_dict_keys = set(old_dict.keys())
    new_dict_keys = set(new_dict.keys())
    intersect_keys = old_dict_keys.intersection(new_dict_keys)
    added = new_dict_keys - old_dict_keys
    removed = old_dict_keys - new_dict_keys
    modified = {o: (old_dict[o], new_dict[o]) for o in intersect_keys if old_dict[o] != new_dict[o]}
    same = set(o for o in intersect_keys if old_dict[o] == new_dict[o] or (old_dict[o] is None and new_dict[o] is None))

    result = {'added': list(added), 'removed': list(removed), 'modified': modified, 'same': list(same)}
    return result


def subset_dict(dict1: dict, dict2: dict):
    return len(set(dict1.items()) & set(dict2.items())) == len(min((dict1, dict2), key=len))


def dict_find_keys(list_dict: list, filter: dict, multiple=True):
    """
    The function searches for dictionary entries in the list of dictionaries.
    :param list_dict: list of dictionaries
    :param filter: dictionary which should be in list_dict
    :param multiple: if True find all entries, else find first entry
    :return:
    """
    result = [x for x in list_dict if subset_dict(filter, x)]
    if not multiple:
        if len(result) > 0:
            result = result[0]
        else:
            result = {}
    return result


def get_args():
    """
    Get dict of args by pairs (e.g. --log_level 'INFO')
    :return: dict of args
    """
    args = sys.argv
    args_d = {}
    for i, arg in enumerate(args):
        if i == 0:
            args_d['name'] = arg
            continue
        if arg.startswith('-') or arg.startswith('--'):
            arg = arg.replace('-', '')
            if i != len(args) - 1:
                if not args[i + 1].startswith('-') or not args[i + 1].startswith('--'):
                    try:
                        args_d[arg] = int(args[i + 1])
                    except ValueError:
                        try:
                            args_d[arg] = float(args[i + 1])
                        except ValueError:
                            args_d[arg] = args[i + 1]
                else:
                    args_d[arg] = True
            else:
                args_d[arg] = True
        else:
            continue

    return args_d


def get_log_param(args: dict):
    """
    Get log_level and log_name from args
    :param args: dict of args from get_args()
    :return: log_level and log_name for set_loglevel()
    """
    log_level = None
    log_name = None
    log_dir = None
    for key in args.keys():
        if key in ['log_level', 'll']:
            log_level = args.get(key)
        if key in ['log_name', 'ln']:
            log_name = args.get(key)
        if key in ['log_dir', 'ld']:
            log_dir = args.get(key)

    if not log_level:
        log_level = 'INFO'
    if not log_name:
        log_name = None
    if not log_dir:
        log_dir = os.path.join(get_base_dir(), 'logs')

    log_params = {'log_level': log_level, 'log_name': log_name, 'log_dir': log_dir}

    return log_params


def get_base_dir():
    try:
        base_dir = os.getcwd()
    except:
        base_dir = ''

    return base_dir


def get_file_extension(filename):
    return ''.join(pathlib.Path(filename).suffixes)


def get_list_files(path: str, filter=None, filter_not_contain: str = None, add_path: bool = False, path2: str = ''):
    """
    Get filelist with path in folder
    :param path: Folder containing files
    :param filter: filter (regexp-like), could be list of str or just str. None by default
    :param filter_not_contain: filter which should not be in filenames (regexp-like)
    :param add_path: bool, will needed add path to file or not
    :param path2: path which need to add
    :return: return list of files with/without path to it
    """
    path = os.path.normpath(path)
    list_files = []
    if add_path:
        if path2 == '':
            path2 = path
        else:
            path2 = os.path.normpath(path2)
    else:
        path2 = ''

    if filter_not_contain:
        filter = r'^((?!{filter_not_contain}).)*$'.format(filter_not_contain=filter_not_contain)
    if filter:
        if type(filter) == list:
            for f in filter:
                list_files.extend([os.path.join(path2, x) for x in os.listdir(path) if re.search(f, x)])
        if type(filter) == str:
            list_files = [os.path.join(path2, x) for x in os.listdir(path) if re.search(filter, x)]
    else:
        list_files = [os.path.join(path2, x) for x in os.listdir(path)]
    
    list_files = list(set(list_files))

    return list_files


def get_list_files_recursive(path: str, filter=None, filter_not_contain: str = None, add_path: bool = False, path2: str = ''):
    """
    Get filelist with path in folder
    :param path: Folder containing files
    :param filter: filter (regexp-like), could be list of str or just str. None by default
    :param filter_not_contain: filter which should not be in filenames (regexp-like)
    :param add_path: bool, will needed add path to file or not
    :param path2: path which need to add
    :return: return list of files with/without path to it
    """
    path = os.path.normpath(path)
    list_files = []
    if add_path:
        if path2 == '':
            path2 = path
        else:
            path2 = os.path.normpath(path2)
    else:
        path2 = ''

    if filter_not_contain:
        filter = r'^((?!{filter_not_contain}).)*$'.format(filter_not_contain=filter_not_contain)

    if filter:
        if type(filter) == list:
            for f in filter:
                for path_, subdirs, files in os.walk(path):
                    list_files.extend([os.path.join(path_, file) for file in files if re.search(f, file)])
        if type(filter) == str:
            for path_, subdirs, files in os.walk(path):
                list_files.extend([os.path.join(path_, file) for file in files if re.search(filter, file)])
    else:
        for path_, subdirs, files in os.walk(path):
            list_files.extend([os.path.join(path_, x) for x in files])

    list_files_ = []
    for file in list_files:
        path_ = os.path.split(file)
        path_ = [os.path.split(path_[0])[1], path_[1]]
        if add_path:
            path_ = [path2, *path_]
        list_files_.append(os.path.join(*path_))

    list_files = list(set(list_files_))

    return list_files


def get_progress(pb_i: int, pb_m: int, pb_p: int=10, process_name: str =''):
    pb_s = pb_m / (100 / pb_p)
    if pb_i == 0:
        logger.get_logger().log_info('{}{:4.0f}% ({:{}.0f} / {})'.format(process_name, 0.0, pb_i, len(str(pb_m)), pb_m))
        # return pb_i
    if math.floor((pb_i + 1) % pb_s) == 0:
        logger.get_logger().log_info('{}{:4.0f}% ({:{}.0f} / {})'.format(process_name, (pb_i + 1) / pb_s * pb_p, (pb_i + 1), len(str(pb_m)), pb_m))
        return (pb_i + 1) / pb_s * pb_p
    if pb_i == (pb_m - 1):
        logger.get_logger().log_info('{}{:4.0f}% ({:{}.0f} / {})'.format(process_name, 100.0, pb_i + 1, len(str(pb_m)), pb_m))
        return (pb_i + 1) / pb_s * pb_p

    return None


def get_double(d: list, key: str):
    D = defaultdict(list)
    d = [x.get(key) for x in d]
    for i, item in enumerate(d):
        D[item].append(i)
    D = {k: v for k, v in D.items() if len(v) > 1}

    return D


def slice_obj(iterable_object, step):
    steps = [i for i in range(len(iterable_object)) if i % step == 0 and i > 0]
    pairs = zip(chain([0], steps), chain(steps, [None]))
    for i, j in pairs:
        yield iterable_object[i:j]


def slice_portion(iterable_obj, portion_qty):
    k, m = divmod(len(iterable_obj), portion_qty)
    return (iterable_obj[i*k+min(i, m):(i+1)*k+min(i+1, m)] for i in range(portion_qty))


def slice_portion_2(iterable_obj, portion_qty):
    return slice_obj(iterable_obj, len(iterable_obj) // portion_qty + 1)


def check_folder_exist(folder_path: str):
    folder_path = os.path.normpath(folder_path)
    if not os.path.exists(folder_path):
        os.makedirs(folder_path)


def simple_translit(text: str):
    symbols = (u"абвгдеёжзийклмнопрстуфхцчшщъыьэюяАБВГДЕЁЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯ",
               u"abvgdeejzijklmnoprstufhzcss_y_euaABVGDEEJZIJKLMNOPRSTUFHZCSS_Y_EUA")
    tr = {ord(a): ord(b) for a, b in zip(*symbols)}
    result = text.translate(tr)
    return result


def varName(var):
    lcls = inspect.stack()[2][0].f_locals
    for name in lcls:
        if id(var) == id(lcls[name]):
            return name
    return None


def get_com(count, word_root, endings=None):
    if endings is None:
        endings = ['', 'а', 'ов']
    if count == 0:
        word = word_root + endings[2]
    else:
        inumber = count % 100
        if inumber >= 11 and inumber <=19:
            word = word_root + endings[2]
        else:
            iinumber = inumber % 10
            if iinumber == 1:
                word = word_root + endings[0]
            elif iinumber == 2 or iinumber == 3 or iinumber == 4:
                word = word_root + endings[1]
            else:
                word = word_root + endings[2]

    return f'{count} {word}'


def parse_date(date_str: Union[bytes, str]):
    date_res = None
    try:
        if re.match(r'\d{4}-\d{2}-\d{2}', date_str):
            date_res = parser.parse(date_str, dayfirst=False)
        elif re.match(r'\d{2}[-.]\d{2}[-.]\d{4}', date_str):
            date_res = parser.parse(date_str, dayfirst=True)
        elif re.match(r'\d{2}/\d{2}/\d{4}', date_str):
            date_res = parser.parse(date_str, dayfirst=False)
    except TypeError:
        return None
    return date_res


def create_json(json_data: Union[str, list, dict, Any], json_fname: str = None, encoding: str = 'utf-8'):
    if not json_fname:
        json_fname = f'{varName(json_data)}.json'
    j = Json(json_fname, encoding)
    j.set_data(json_data)
    j.create()


def parse_json(json_fname: str, encoding: str = 'utf-8', return_: Union[str, list, dict] = None):
    j = Json(json_fname, encoding, return_)
    return j.parse()
    

def create_yaml(yaml_data: Union[str, list, dict, Any], yaml_fname: str = None, encoding: str = 'utf-8'):
    if not yaml_fname:
        yaml_fname = f'{varName(yaml_data)}.yml'
    y = Yaml(yaml_fname, encoding)
    y.set_data(yaml_data)
    y.create()


def parse_yaml(yaml_fname: str, encoding: str = 'utf-8', return_: Union[str, list, dict] = None):
    y = Yaml(yaml_fname, encoding, return_)
    return y.parse()


class BaseMl:
    def __init__(self, fname: Optional[str] = None, encoding: str = 'utf-8', return_: Union[str, list, dict] = None):
        if fname:
            file_dir, file_name = os.path.split(fname)
        else:
            file_dir, file_name = None, None
        self.fname = file_name
        self.fpath = file_dir
        self.encoding = encoding
        self.return_ = return_
        self.data = None

    def __repr__(self):
        return f'{self.__cls__}({self.fname}, {self.encoding}, {self.return_}) with {self.data}'

    def reinit(self, fname: Optional[str] = None, encoding: str = 'utf-8', return_: Union[str, list, dict] = None):
        if fname:
            file_dir, file_name = os.path.split(fname)
        else:
            file_dir, file_name = None, None
        self.fname = file_name
        self.fpath = file_dir
        self.encoding = encoding
        self.return_ = return_

    def set_data(self, data: Union[str, list, dict, Any]):
        self.data = data

    def get_data(self):
        return self.data


class Yaml(BaseMl):
    def parse(self):
        try:
            with open(os.path.join(self.fpath, self.fname), 'r', encoding=self.encoding) as yaml_file:
                self.data = yaml.full_load(yaml_file)
        except:
            self.data = self.return_
        return self.data

    def create(self):
        if self.fname:
            check_folder_exist(self.fpath)
            with open(os.path.join(self.fpath, self.fname), 'w', encoding=self.encoding) as outfile:
                yaml.dump(self.data, outfile, default_flow_style=False, indent=4,
                          allow_unicode=True, sort_keys=False)
        else:
            raise NoFileNamePassedError('Please provide filename first!')


class Json(BaseMl):
    def parse(self):
        try:
            with open(os.path.join(self.fpath, self.fname), 'r', encoding=self.encoding) as json_file:
                json_str = json_file.read()
                self.data = json.loads(json_str, object_hook=self.date_hook)
        except:
            self.data = self.return_
        return self.data

    def create(self):
        if self.fname:
            check_folder_exist(self.fpath)
            dthandler = lambda obj: obj.isoformat() if isinstance(obj, datetime) or isinstance(obj, date) else None

            with open(os.path.join(self.fpath, self.fname), 'w', encoding=self.encoding) as outfile:
                json.dump(self.data, outfile, ensure_ascii=False, default=dthandler, indent=4)
        else:
            raise NoFileNamePassedError('Please provide filename first!')

    def dumps(self):
        dthandler = lambda obj: obj.isoformat() if isinstance(obj, datetime) or isinstance(obj, date) else None
        response = json.dumps(self.data, ensure_ascii=False, default=dthandler, indent=4)
        return response

    @staticmethod
    def date_hook(json_dict: dict):
        """
        Hook which used for translate date string in json file to datetime type
        :param json_dict: dict from json file
        :return: dict from json after translating
        """
        for (key, value) in json_dict.items():
            date_value = parse_date(value)
            if isinstance(date_value, (date, datetime)):
                json_dict[key] = date_value
        return json_dict


def pickle_data(data: Union[Any], fname: str):
    with open(fname, 'wb') as f:
        pickle.dump(data, f)


def unpickle_data(fname: str, return_: Union[str, list, dict] = None):
    try:
        with open(fname, 'rb') as f:
            data = pickle.load(f)
    except FileNotFoundError:
        data = return_
    return data


if __name__ == '__main__':
    pass
